/*
 * A page mapper providing copied memory pages or deferring to another page
 * mapper to satisfy file accesses.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/dataspace.h>

#include <string.h>

#include "copied_page_mapper.h"



/* Provide mapped pages populated with the given 'mapper', with pages obtained
   from the given 'pages' collection. */

CopiedPageMapper::CopiedPageMapper(GenericPageMapper *mapper)
: _mapper(mapper)
{
    _size = _mapper->get_data_size();
}

/* Upon deallocation, release allocated flexpages. */

CopiedPageMapper::~CopiedPageMapper()
{
    while (!_queue.empty())
    {
        Flexpage *flexpage = _queue.front();

        _queue.pop_front();
        _memory.release(flexpage->region);
        delete flexpage;
    }
}

/* Interface for the pager. */

/* Return a flexpage providing access to the indicated file 'offset'.

   The returned flexpage will either provide direct access to the underlying
   file content or be a completely new flexpage modified through writes to
   memory. */

Flexpage *CopiedPageMapper::get(offset_t offset, map_flags_t flags)
{
    Flexpage *f = find(offset);

    /* Defer to the underlying page mapper if no flexpage was found. */

    if (f == NULL)
    {
        f = _mapper->get(offset, flags);

        /* Replicate the underlying flexpage if writing. */

        if (flags & L4RE_DS_F_W)
        {
            Flexpage *rf = replicate_flexpage(offset, f);
            _mapper->queue(f);
            f = rf;
        }
    }

    /* Upgrade the access flags. */

    f->upgrade(flags);
    return f;
}

/* Queue the given 'flexpage'. */

void CopiedPageMapper::queue(Flexpage *flexpage)
{
    if (dynamic_cast<CopiedFlexpage *>(flexpage) == NULL)
        _mapper->queue(flexpage);
}

/* Flush pages in the given range from 'start' with 'size'. */

void CopiedPageMapper::flush_all(offset_t start, offset_t size)
{
    /* NOTE: This might be superfluous since copy-on-write regions should not
             update content. */

    _mapper->flush_all(start, size);
}

/* Return the maximum extent of the mapped resource. */

offset_t CopiedPageMapper::get_data_size()
{
    return _size;
}

/* Set the maximum extent of the mapped resource. */

void CopiedPageMapper::set_data_size(offset_t size)
{
    _size = size;
}

/* Internal flexpage retrieval methods. */

/* Find an existing flexpage for 'offset'. */

Flexpage *CopiedPageMapper::find(offset_t offset)
{
    return _map.find(offset);
}

/* Replicate an underlying flexpage for the file 'offset' using the given
   'flexpage'. */

Flexpage *CopiedPageMapper::replicate_flexpage(offset_t offset, Flexpage *flexpage)
{
    /* Obtain a new memory region and a new flexpage. */

    Region *region = _memory.region(flexpage->size);
    Flexpage *replicated = new CopiedFlexpage(region);

    /* Configure the flexpage for the file region. */

    replicated->reset(offset);

    /* Copy the contents of the underlying flexpage. */

    memcpy((void *) region->start, (const void *) flexpage->region->start, flexpage->size);

    /* Register the flexpage for future accesses and for deallocation. */

    _map.insert(replicated);
    _queue.push_back(replicated);
    return replicated;
}



/* Copied flexpage constructor. */

CopiedFlexpage::CopiedFlexpage(Region *region)
: Flexpage(region)
{
}

// vim: tabstop=4 expandtab shiftwidth=4
